
structure Types =
struct
    
    type var = string

    datatype typ =
        UNIT
      | INT
      | ARROW of typ * typ
      | VAR of var
		| ALL of var * typ
		| BOOL
    
    exception TypeError of string

    (* type equality is tricker now since we have alpha-equivalent
       types which are not equal by the ML idea of datatype equality *)

    fun empty msg v = raise TypeError (msg ^ v)
    fun ident v = VAR v

    infix ++
    fun (G ++ (v, e)) vv = if vv = v then e else G vv

	 fun typeok D (ARROW(t1,t2)) = typeok D t1 andalso typeok D t2
		| typeok D (VAR v) = D v
		| typeok D (ALL (v, t)) = typeok (D ++ (v, true)) t
		| typeok D _ = true

    fun tsubst _ INT = INT
      | tsubst _ UNIT = UNIT
		| tsubst _ BOOL = BOOL
      | tsubst S (VAR v) = S v
      | tsubst S (ARROW (t1, t2)) = ARROW(tsubst S t1, tsubst S t2)
      | tsubst S (ALL (v, t)) = ALL(v, tsubst (S ++ (v, VAR v)) t)

	 local val ctr = ref 0
	 in fun newvar () = Int.toString (ctr := !ctr + 1; !ctr)
	 end

    fun typeeq (t1, t2) =
        let
            fun te (INT, INT) = true
              | te (UNIT, UNIT) = true
				  | te (BOOL, BOOL) = true
              | te (ARROW (a1, a2), ARROW (b1, b2)) = te (a1, b1) andalso te (a2, b2)
              | te (VAR v, VAR w) = v = w
              | te (ALL (v1, t1), ALL (v2, t2)) =
					 let val mv = VAR(newvar ())
					 in te (tsubst (ident ++ (v1, mv)) t1, tsubst (ident ++ (v2, mv)) t2)
					 end
              | te _ = false
        in
            te (t1, t2)
        end

    fun typetostring INT = "INT"
      | typetostring (ARROW(t1,t2)) = "(" ^ typetostring t1 ^ " -> " ^ typetostring t2 ^ ")"
      | typetostring (ALL(v,t)) = "(ALL " ^ v ^ ". " ^ typetostring t ^ ")"
      | typetostring (VAR v) = v
      | typetostring UNIT = "UNIT"
		| typetostring BOOL = "BOOL"

end